from ast import literal_eval
import json

from ipywidgets import register, CallbackDispatcher, DOMWidget
from traitlets import Any, Bool, Int, Unicode

from ..data_utils.binary_transfer import data_buffer_serialization
from ._frontend import module_name, module_version
from .debounce import debounce


def store_selection(widget_instance, payload):
    """Callback for storing data on click"""
    try:
        if payload.get("data") and payload["data"].get("object"):
            datum = payload["data"]["object"]
            widget_instance.selected_data.append(datum)
        else:
            widget_instance.selected_data = []
    except Exception as e:
        widget_instance.handler_exception = e


@register
class DeckGLWidget(DOMWidget):
    """
    Jupyter environment widget that takes JSON and
    renders a deck.gl visualization based on provided properties.

    You may set a Mapbox API key as an environment variable to use Mapbox maps in your visualization

    Attributes
    ----------
        json_input : str, default ''
            JSON as a string meant for reading into deck.gl JSON API
        mapbox_key : str, default ''
            API key for Mapbox map tiles
        height : int, default 500
            Height of Jupyter notebook cell, in pixels
        width : int or str, default "100%"
            Width of Jupyter notebook cell, in pixels or, if a string, a CSS width
        tooltip : bool or dict of {str: str}, default True
            See the ``Deck`` constructor.
        google_maps_key : str, default ''
            API key for Google Maps
        selected_data : list of dict, default []
            Data selected on click, if the pydeck Jupyter widget is enabled for server use
    """

    _model_name = Unicode("JupyterTransportModel").tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode("JupyterTransportView").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    carto_key = Unicode("", allow_none=True).tag(sync=True)
    mapbox_key = Unicode("", allow_none=True).tag(sync=True)
    google_maps_key = Unicode("", allow_none=True).tag(sync=True)

    json_input = Unicode("").tag(sync=True)
    data_buffer = Any(default_value=None, allow_none=True).tag(sync=True, **data_buffer_serialization)
    custom_libraries = Any(allow_none=True).tag(sync=True)
    configuration = Any(allow_none=True).tag(sync=True)
    tooltip = Any(True).tag(sync=True)
    height = Int(500).tag(sync=True)
    width = Any("100%").tag(sync=True)

    def __init__(self, **kwargs):
        super(DeckGLWidget, self).__init__(**kwargs)
        self._hover_handlers = CallbackDispatcher()
        self._click_handlers = CallbackDispatcher()
        self._resize_handlers = CallbackDispatcher()
        self._view_state_handlers = CallbackDispatcher()
        self._drag_handlers = CallbackDispatcher()
        self._drag_start_handlers = CallbackDispatcher()
        self._drag_end_handlers = CallbackDispatcher()
        self.on_msg(self._handle_custom_msgs)

        self.handler_exception = None
        self.selected_data = []
        self.on_click(store_selection)

    def on_hover(self, callback, remove=False):
        self._hover_handlers.register_callback(callback, remove=remove)

    def on_resize(self, callback, remove=False):
        self._resize_handlers.register_callback(callback, remove=remove)

    def on_view_state_change(self, callback, debounce_seconds=0.2, remove=False):
        callback = debounce(debounce_seconds)(callback) if debounce_seconds > 0 else callback
        self._view_state_handlers.register_callback(callback, remove=remove)

    def on_click(self, callback, remove=False):
        self._click_handlers.register_callback(callback, remove=remove)

    def on_drag_start(self, callback, remove=False):
        self._drag_start_handlers.register_callback(callback, remove=remove)

    def on_drag(self, callback, remove=False):
        self._drag_handlers.register_callback(callback, remove=remove)

    def on_drag_end(self, callback, remove=False):
        self._drag_end_handlers.register_callback(callback, remove=remove)

    def _handle_custom_msgs(self, _, content, buffers=None):
        content = json.loads(content)
        event_type = content.get("type", "")
        if event_type == "deck-hover-event":
            self._hover_handlers(self, content)
        elif event_type == "deck-resize-event":
            self._resize_handlers(self, content)
        elif event_type == "deck-view-state-change-event":
            self._view_state_handlers(self, content)
        elif event_type == "deck-click-event":
            self._click_handlers(self, content)
        elif event_type == "deck-drag-start-event":
            self._drag_start_handlers(self, content)
        elif event_type == "deck-drag-event":
            self._drag_handlers(self, content)
        elif event_type == "deck-drag-end-event":
            self._drag_end_handlers(self, content)



from functools import total_ordering

from .base import PydeckType


@total_ordering
class Function(PydeckType):
    """Indicate a function type with arguments and set already in pydeck

    Parameters
    ----------

    name : str
        Function name
    **kwargs
        arguments and value of each argument to be storing the function information
    """

    __KEY = "@@function"

    def __init__(self, name: str, **kwargs):
        self.name = name
        self.arguments = kwargs

    def __lt__(self, other):
        return str(self) < str(other)

    def __eq__(self, other):
        return str(self) == str(other)

    def serialize(self):
        repr = {self.__KEY: self.name}
        repr.update(self.arguments)
        return repr



import warnings


DARK = "dark"
LIGHT = "light"
SATELLITE = "satellite"
ROAD = "road"
DARK_NO_LABELS = "dark_no_labels"
LIGHT_NO_LABELS = "light_no_labels"

MAPBOX_LIGHT = "mapbox://styles/mapbox/light-v9"
MAPBOX_DARK = "mapbox://styles/mapbox/dark-v9"
MAPBOX_ROAD = "mapbox://styles/mapbox/streets-v9"
MAPBOX_SATELLITE = "mapbox://styles/mapbox/satellite-v9"

CARTO_DARK = "https://basemaps.cartocdn.com/gl/dark-matter-gl-style/style.json"
CARTO_DARK_NO_LABELS = "https://basemaps.cartocdn.com/gl/dark-matter-nolabels-gl-style/style.json"
CARTO_LIGHT = "https://basemaps.cartocdn.com/gl/positron-gl-style/style.json"
CARTO_LIGHT_NO_LABELS = "https://basemaps.cartocdn.com/gl/positron-nolabels-gl-style/style.json"
CARTO_ROAD = "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json"

GOOGLE_SATELLITE = "satellite"
GOOGLE_ROAD = "roadmap"

styles = {
    DARK: {"mapbox": MAPBOX_DARK, "carto": CARTO_DARK},
    DARK_NO_LABELS: {"carto": CARTO_DARK_NO_LABELS},
    LIGHT: {"mapbox": MAPBOX_LIGHT, "carto": CARTO_LIGHT},
    LIGHT_NO_LABELS: {"carto": CARTO_LIGHT_NO_LABELS},
    ROAD: {"carto": CARTO_ROAD, "google_maps": GOOGLE_ROAD, "mapbox": MAPBOX_ROAD},
    SATELLITE: {"mapbox": MAPBOX_SATELLITE, "google_maps": GOOGLE_SATELLITE},
}


def get_from_map_identifier(map_identifier: str, provider: str) -> str:
    """Attempt to get a style URI by map provider, otherwise pass the map identifier
    to the API service

    Provide reasonable cross-provider default map styles

    Parameters
    ----------
    map_identifier : str
        Either a specific map provider style or a token indicating a map style. Currently
        tokens are "dark", "light", "satellite", "road", "dark_no_labels", or "light_no_labels".
        Not all map styles are available for all providers.
    provider : str
        One of "carto", "mapbox", or "google_maps", indicating the associated base map tile provider.

    Returns
    -------
    str
        Base map URI

    """
    try:
        return styles[map_identifier][provider]
    except KeyError:
        return map_identifier



from .viewport_helpers import compute_view  # noqa
from .type_checking import has_geo_interface, is_pandas_df, records_from_geo_interface  # noqa
from .color_scales import assign_random_colors  # noqa



def is_pandas_df(obj):
    """Check if an object is a Pandas DataFrame

    Returns
    -------
    bool
        Returns True if object is a Pandas DataFrame and False otherwise
    """
    return obj.__class__.__module__ == "pandas.core.frame" and obj.to_records and obj.to_dict


def has_geo_interface(obj):
    return hasattr(obj, "__geo_interface__")


def records_from_geo_interface(data):
    """Un-nest data from object implementing __geo_interface__ standard"""
    flattened_records = []
    for d in data.__geo_interface__.get("features"):
        record = d.get("properties", {})
        geom = d.get("geometry", {})
        record["geometry"] = geom
        flattened_records.append(record)
    return flattened_records
